(***********************************************************************)
(*                                                                     *)
(*                       Banana Algebra Code                           *)
(*                                                                     *)
(*                         Claus Brabrand                              *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2005 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)


open Ast
open String
open Misc
open List

let file = ref stdout  

(* -- ast -------------------------------------------------- *)

let rec
code_ast (ps : prod list) =
  begin
    fprintln !file "type ast = " ;
    fprintln !file "    Ast of int * ast list" ;
    fprintln !file "  | Term of string" ;
    fprintln !file "  | Gap of string" ;
  end

(* -- regexp -------------------------------------------------- *)

and
code_regexp = function
    Union(r1,r2,_)        -> begin
                               fprint !file "( " ;
                               code_regexp r1 ;
                               fprint !file " | " ;
                               code_regexp r2 ;
                               fprint !file " )"
                             end
  | Concat(r1,r2,_)       -> begin
                               fprint !file "( " ;
                               code_regexp r1 ;
                               fprint !file " " ;
                               code_regexp r2 ;
                               fprint !file " )"
                             end
  | Optional(r,_)        ->  begin
                               fprint !file "( " ;
                               code_regexp r ;
                               fprint !file " )?"
                             end
  | Star(r,_)            ->  begin
                               fprint !file "( " ;
                               code_regexp r ;
                               fprint !file " )*"
                             end
  | Plus(r,_)             -> begin
                               fprint !file "( " ;
                               code_regexp r ;
                               fprint !file " )+"
                             end
  | CharClass(l,_)        -> begin
                               fprint !file "[ " ;
                               code_charclasses l ;
                               fprint !file " ]"
                             end
  | NegCharClass(l,_)     -> begin
                               fprint !file "[^ " ;
                               code_charclasses l ;
                               fprint !file " ]"
                             end
  | Dot(_)                -> fprint !file "_"
  | String(s,_)           -> fprint !file ("\"" ^ s ^ "\"")
  | RegexpID(v,_)         -> fprint !file ("t_" ^ v)
  | Ch(c,_)               -> code_charexp c

and
code_charclasses = function
    [] -> ()
  | c::cc -> begin
               code_charclass c ;
               if (cc != []) then fprint !file " " ;
               code_charclasses cc
             end

and
code_charclass = function
    Range(c1,c2,_) -> begin
                        code_charexp c1 ;
                        fprint !file " - " ;
                        code_charexp c2
                      end
  | CharExp(c,_)   -> code_charexp c

and
code_charexp = function
    Char(c,_)    -> fprint !file ("'" ^ c ^ "'")
  | EscChar(c,_) -> fprint !file ("'" ^ c ^ "'")

(* -- lexer -------------------------------------------------- *)

and
extract_keywords_entities es = match es with
    []    -> []
  | Es(s,_)::es -> [s] @ (extract_keywords_entities es)
  | Et(t,_)::es -> (extract_keywords_entities es)
  | En(n,_)::es -> (extract_keywords_entities es)

and
extract_keywords_prod (n,p,es,_) = extract_keywords_entities es

and
extract_keywords lprods kset = match lprods with
    []      -> kset
  | lp::lps -> let kset' = (usort compare (extract_keywords_prod lp))
               in extract_keywords lps (umerge compare kset kset')

and
code_lexer_term_def (t,r,_) =
  begin
    fprint !file ("let t_" ^ t ^ " = ") ;
    code_regexp r ;
    fnewline !file
  end

and
code_lexer_term_use (t,_,_) =
  begin
    fprintln !file ("| t_" ^ t ^ " { T_" ^ t ^ "(Lexing.lexeme lexbuf) }") ;
  end

and
code_lexer_start n = 
  begin
    fprintln !file ("| \"<<<" ^ n ^ ">>>\" { S_" ^ n ^ " }") ;
  end

and
code_lexer_inline_keyword k i =
  begin
    fprintln !file ( "| \"" ^ k ^ "\" { T_" ^ (int_to_string i) ^ " }" )
  end

and
code_lexer_inline_keywords kset i = match kset with
    [] -> ()
  | k::kset -> begin
                 code_lexer_inline_keyword k i ;
                 code_lexer_inline_keywords kset (i+1)
               end

and
code_lexer (ts : term list) (ps : prod list) (kset : string list) = 
  begin
    fprintln !file "{ open Parser" ;
    fprintln !file "  open Lexing " ;
    fprintln !file "" ;
    fprintln !file "  exception LexError of string * int * int" ;
    fnewline !file ;
    fprintln !file "}" ;
    fnewline !file ;
    iter code_lexer_term_def ts ;
    fnewline !file ;
    fprintln !file "rule token = parse" ;
    fprintln !file ("| eof { EOF }") ;
    fprintln !file ("| ['\\t' '\\r' '\\n' ' ']+ { token lexbuf }") ;
    code_lexer_inline_keywords kset 0 ;
    iter code_lexer_term_use ts ;
    iter code_lexer_start (nts ps) ;
    fprintln !file ("| _ { raise (LexError(Lexing.lexeme lexbuf, Lexing.lexeme_start lexbuf, Lexing.lexeme_end lexbuf )) }") ;
  end

(* -- parser -------------------------------------------------- *)

let rec
code_parser_declare_keyword k i =
  begin
    fprint !file ( " T_" ^ (int_to_string i) )
  end

and
code_parser_declare_keywords kset i = match kset with
    [] -> ()
  | k::kset -> begin
                 code_parser_declare_keyword k i ;
                 code_parser_declare_keywords kset (i+1)
               end

and
code_parser_declare_term (t,_,_) = 
  begin
    fprint !file ( " T_" ^ t )
  end

and
code_parser_declare_start n = 
  begin
    fprint !file ( " S_" ^ n )
  end

and
code_parser_declare_tokens ts nts kset = 
  begin
    fprint !file "%token EOF" ;
    code_parser_declare_keywords kset 0 ;
    iter code_parser_declare_start nts ;
    fnewline !file ;
    fnewline !file ;
    fprint !file "%token <string>" ;
    iter code_parser_declare_term ts ;
    fnewline !file
  end

and
code_parser_start n = fprint !file (" " ^ n);

and
code_parser_starts nts = match nts with
  [] -> ()
 | n::nts -> begin
               fprintln !file ("S_" ^ n ^ " n_" ^ n ^ " { $2 }") ;
               if (nts <> []) then fprint !file "| " ;
               code_parser_starts nts
             end

and
code_parser_entity kset e = match e with
    Es(s,_) -> fprint !file (" T_" ^ int_to_string (index_of kset s))
  | Et(t,_) -> fprint !file (" T_" ^ t)
  | En(n,_) -> fprint !file (" n_" ^ n)

and
code_parser_entity_bodies es inx delim = match es with
  [] -> ()
  | Es(s,_)::es -> code_parser_entity_bodies es (inx+1) delim
  | Et(t,_)::es -> begin
                     fprint !file (delim ^ "Term($" ^ (int_to_string inx) ^ ")" ) ;
                     code_parser_entity_bodies es (inx+1) ";"
                   end
  | En(n,_)::es -> begin
                     fprint !file (delim ^ "$" ^ (int_to_string inx)) ;
                     code_parser_entity_bodies es (inx+1) ";"
                   end

and
code_parser_prods (ps : prod list) kset last_n p_num = match ps with
   [] -> ()
 | (n,p,es,_)::ps  -> begin
                        if (n <> last_n) (* new nonterminal n *)
                        then fprintln !file ("\nn_" ^ n ^ "\n: S_" ^ n ^ " { Gap(\"" ^ n ^ "\") }" ) ;
                        fprint !file "|" ;
                        iter (code_parser_entity kset) es ;
                        fprint !file (" { Ast (" ^ (int_to_string p_num) ^ ",[" ) ;
                        code_parser_entity_bodies es 1 "" ;
                        fprintln !file "]) }" ;
                        code_parser_prods ps kset n (p_num+1)
                      end

and
code_parser (ts : term list) (ps : prod list) (kset : string list) = 
  begin
    fprintln !file "\n%{" ;
    fprintln !file "  open Ast" ;
    fprintln !file "  open List" ;
    fprintln !file "  open Printf" ;
    fprintln !file "%}\n" ;
    code_parser_declare_tokens ts (nts ps) kset ;
    fprintln !file "" ;
    fprintln !file "%start start" ;
    fprintln !file "%type <Ast.ast> start" ;
    fprintln !file "\n%%\n" ;
    fprint !file "start \n: " ;
    code_parser_starts (nts ps) ;
    fprintln !file "\n" ;
    code_parser_prods ps kset "" 0 ;
    fprintln !file "\n%%\n" ;
    fnewline !file ;
  end


(* -- pretty -------------------------------------------------- *)

and
code_pretty_entities es inx = match es with
    [] -> ()
  | Es(s,_)::es -> begin
                     fprintln !file ("          Printf.printf \"%s\" \"" ^ s ^ " \";") ; 
                     code_pretty_entities es inx
                   end
  | Et(t,_)::es -> begin
                     fprintln !file ("          pretty (nth asts " ^ (int_to_string inx) ^ "); (* $" ^ t ^ " *)") ; 
                     code_pretty_entities es (inx+1)
                   end
  | En(n,_)::es -> begin
                     fprintln !file ("          pretty (nth asts " ^ (int_to_string inx) ^ "); (* " ^ n ^ " *)") ; 
                     code_pretty_entities es (inx+1)
                   end

and
code_pretty_prods ps inx = match ps with
    [] -> ()
  | (n,p,es,_)::ps -> begin
                        fprintln !file ("    | " ^ (int_to_string inx) ^ " ->" ) ;
                        fprintln !file ("        begin") ;
                        code_pretty_entities es 0 ;
                        fprintln !file ("        end") ;
                        code_pretty_prods ps (inx+1)
                      end

and
code_pretty (ps : prod list) = 
  begin
    fprintln !file "let rec" ;
    fprintln !file "pretty = function" ;
    fprintln !file "    Term(t)     -> Printf.printf \"%s\" (t ^ \" \")" ;
    fprintln !file "  | Gap(nt)     -> Printf.printf \"%s\" (\"<<<\" ^ nt ^ \">>> \")" ;
    fprintln !file "  | Ast(i,asts) -> match i with" ;
    code_pretty_prods ps 0 ;
    fprintln !file "    | _ -> raise (FatalError(\"Illegal AST\"))" ;
    fprintln !file ""
  end

(* -- printast -------------------------------------------------- *)

and
code_printast_entities es inx = match es with
    [] -> ()
  | Es(s,_)::es -> begin
                     code_printast_entities es inx
                   end
  | Et(t,_)::es -> begin
                     if (inx > 0) then fprintln !file ("          Printf.printf \",\" ;") ;
                     fprintln !file ("          printast (nth asts " ^ (int_to_string inx) ^ "); (* $" ^ t ^ " *)") ; 
                     code_printast_entities es (inx+1)
                   end
  | En(n,_)::es -> begin
                     if (inx > 0) then fprintln !file ("          Printf.printf \",\" ;") ;
                     fprintln !file ("          printast (nth asts " ^ (int_to_string inx) ^ "); (* " ^ n ^ " *)") ; 
                     code_printast_entities es (inx+1)
                   end

and
code_printast_prods ps inx = match ps with
    [] -> ()
  | (n,p,es,_)::ps -> begin
                        fprintln !file ("    | " ^ (int_to_string inx) ^ " ->" ) ;
                        fprintln !file ("        begin") ;
                        fprintln !file ("          Printf.printf \"%s\" \"" ^ (n ^ "." ^ p) ^ "(\" ;") ;
                        code_printast_entities es 0 ;
                        fprintln !file ("          Printf.printf \")\" ;") ;
                        fprintln !file ("        end") ;
                        code_printast_prods ps (inx+1)
                      end

and
code_printast (ps : prod list) = 
  begin
    fprintln !file "let rec" ;
    fprintln !file "printast = function" ;
    fprintln !file "    Term(t)     -> Printf.printf \"%s\" (\"'\" ^ t ^ \"'\")" ;
    fprintln !file "  | Gap(nt)     -> Printf.printf \"%s\" (\"<<<\" ^ nt ^ \">>>\")" ;
    fprintln !file "  | Ast(i,asts) -> match i with" ;
    code_printast_prods ps 0 ;
    fprintln !file "    | _ -> raise (FatalError(\"Illegal AST\"))" ;
    fprintln !file ""
  end

(* -- transform -------------------------------------------------- *)

and
code_transform_asts asts lt = match asts with
    [] -> ()
  | ast::asts -> begin
                   code_transform_ast ast lt ;
                   if (asts <> []) then fprint !file ("; ") ;
                   code_transform_asts asts lt
                 end

and
code_transform_ast ast lt = match ast with
  | Aast(n,p,asts,_) -> begin
                          let tgt_inx = (index_of (snd lt) (find (function (_n,_p,_es,_pos) -> ((n = _n && p = _p))) (snd lt))) (* TODO: simplify this index calculation *)
                          in fprint !file ("Ast(" ^ (int_to_string tgt_inx) ^ ",[") ;
                          code_transform_asts asts lt ;
                          fprint !file ("])") ;
                        end
  | Aarg(k,inx,_)    -> fprint !file ("(transform (nth asts " ^ (int_to_string inx) ^ "))")
  | Aterm(t,s,_)     -> fprint !file ("Term(\"" ^ s ^ "\")")

and
code_transform_xprods xps lt inx = match xps with
    [] -> ()
  | (n,p,ast,_)::xps -> begin
                          fprint !file ("    | " ^ (int_to_string inx) ^ " -> ") ;
                          code_transform_ast ast lt ;
                          fprintln !file (" (* " ^ n ^ "." ^ p ^ " *)") ;
                          code_transform_xprods xps lt (inx+1)
                        end

and
code_transform xps lt =
  begin
    fprintln !file "let rec" ;
    fprintln !file "transform = function" ;
    fprintln !file "    Term(t)     -> Term(t)" ;
    fprintln !file "  | Gap(nt)     -> Gap(nt)" ;
    fprintln !file "  | Ast(i,asts) -> match i with" ;
    code_transform_xprods xps lt 0 ;
    fprintln !file "    | _ -> raise (FatalError(\"Illegal AST\"))" ;
    fprintln !file ""
  end

(* -- main -------------------------------------------------- *)

and
code_main () = 
  begin
    fprintln !file "let main =" ;
    fprintln !file "  begin" ;
    fprintln !file "    let lexbuf = Lexing.from_channel stdin" ;
    fprintln !file "    in" ;
    fprintln !file "      try " ;
    fprintln !file "        let ast = Parser.start Lexer.token lexbuf" ;
    fprintln !file "        in (pretty (transform ast)) ;" ;
    fprintln !file "        Printf.printf \"\\n\"" ;
    fprintln !file "      with " ;
    fprintln !file "        Lexer.LexError(s,x,y) ->" ;
    fprintln !file "          Printf.fprintf stderr \"%s\" (\"*** lexical error at position \" ^ (Printf.sprintf \"%d\" x) ^ \"--\" ^ (Printf.sprintf \"%d\" y) ^ \": unknown character '\" ^ s ^ \"'\")" ;
    fprintln !file "      | Parsing.Parse_error ->" ;
    fprintln !file "          Printf.fprintf stderr \"%s\" (\"*** syntax error at position \" ^ (Printf.sprintf \"%d\" (Lexing.lexeme_start lexbuf)) ^ \"--\" ^ (Printf.sprintf \"%d\" (Lexing.lexeme_end lexbuf)) ^ \": error at '\" ^ (Lexing.lexeme lexbuf) ^ \"'\")" ;
    fprintln !file "  end" ;
  end


(* -- code -------------------------------------------------- *)

and
code (xprods,tau,ls,lt) filename = 
  begin
    let kset = extract_keywords (snd ls) []
    in 
    (* code ast *)
    file := open_out ("ast.mli") ;
    code_ast (snd ls) ;
    close_out !file ;

    (* code lexer *)
    file := open_out ("lexer.mll") ;
    code_lexer (fst ls) (snd ls) kset ;
    close_out !file ;

    (* code parser *)
    file := open_out ("parser.mly") ;
    code_parser (fst ls) (snd ls) kset ;
    close_out !file ;

    (* code main *)
    file := open_out ("main.ml") ;
    fprintln !file "open Ast" ;
    fprintln !file "open List" ;
    fprintln !file "" ;
    fprintln !file "exception ParseError of string" ;
    fprintln !file "exception FatalError of string" ;
    fprintln !file "" ;
    code_pretty (snd lt) ;
    code_printast (snd ls) ;
    code_transform xprods lt ;
    code_main () ;
    close_out !file
  end
