(***********************************************************************)
(*                                                                     *)
(*                   Banana Algebra Inline parser                      *)
(*                                                                     *)
(*                         Jacob Andersen                              *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)


open Ast
open CodeXML

let xsgFilename = ref ""
let fragmentFilename = ref ""
let lastLang : lcon ref = ref ((Empty(Lexing.dummy_pos,Lexing.dummy_pos),
                                (Lexing.dummy_pos,Lexing.dummy_pos)),[],[])
let dPos = (Lexing.dummy_pos, Lexing.dummy_pos)

let
check_num n =
  try
    let i = int_of_string n in
    if i>0 then i else raise (Failure "")
  with Failure _ -> raise (Invalid_argument ("Invalid gap-number '" ^ n ^ "'"))

let
check_typ n t tenv =
  let i = check_num n in
  let t' = try List.nth tenv (i-1) with Failure _ -> "" in
  match t, t' with
      "", "" -> raise (Invalid_argument ("Type of gap number " ^
                       (string_of_int i) ^ " unknown. Explicit type needed"))
    | "", t  -> t
    | t, ""  -> t
    | t, t'  -> if t=t' then t else raise (Invalid_argument (
                  "Type conflict at gap number " ^ (string_of_int i) ^
                  ". Implicit type: '" ^ t' ^ "', given type: '" ^ t ^ "'"))
          

let 
modify_string s = 
  begin
    let s' = String.make ((String.length s)*2) 'c' in
    let i = ref 1 in
    String.iter (fun c -> s'.[!i] <- c ; i:=!i+2) s ;
    s'
  end

let rec
modify_regexp = function
    Union (r1,r2,p)  -> Union(modify_regexp r1, modify_regexp r2,p)
  | Concat (r1,r2,p) -> Concat(modify_regexp r1, modify_regexp r2,p)
  | Optional(r,p)    -> Optional(modify_regexp r,p)
  | Star(r,p)        -> Star(modify_regexp r,p)
  | Plus(r,p)        -> Plus(modify_regexp r,p)
  | CharClass(c,p)   -> Concat(Ch('c',p),CharClass(c,p),p)
  | NegCharClass(c,p)-> Concat(Ch('c',p),NegCharClass(c,p),p)
  | Dot p            -> Concat(Ch('c',p),Dot p,p)
  | String(s,p)      -> String(modify_string s,p)
  | RegexpID(v,p)    -> RegexpID(v,p)
  | Ch(c,p)          -> Concat(Ch('c',p),Ch(c,p),p)
  | Empty p          -> Empty p

let 
modify_entity = function
    Es(s,p) -> Es(modify_string s,p)
  | En(v,p) -> En(v,p)

let rec
gapRules ps prevN = match ps with
    [] -> []
  | (n,_,_,_)::ps ->
      if prevN <> n then
        (n,".",[Es("("^n^":",dPos);En(".",dPos);Es(")",dPos)],dPos)
          :: gapRules ps n
      else
        gapRules ps n


let
inline_parse (tenv:string list) (lcon:lcon) (n:var) (s:string) : ast = 
  begin

    (* Create files if not exists *)

    if (!xsgFilename = "") then
      begin
        let (n,f) = Filename.open_temp_file "shape" ".xsg" in
        xsgFilename := n ;
        close_out f ;
        let (n,f) = Filename.open_temp_file "shape" ".txt" in
        fragmentFilename := n ;
        close_out f ;
        at_exit (fun () -> Sys.remove !xsgFilename ;
                           Sys.remove !fragmentFilename)
      end ;

    (* Generate new XSugar file *)

    if lcon != !lastLang then 
      begin

        (* Modify grammar for gap-preprocessing *)

        lastLang := lcon ;

        let ((w,wpos),ts,ps) = lcon in
        let lcon' = 
          ((modify_regexp w,wpos),
           (".",Plus(CharClass([Range('0','9',dPos)],dPos),dPos),dPos)::
             List.map
              (fun(v,r,p) -> (v,Union(modify_regexp r,
                                      Concat(Concat(Concat(Concat(Ch('(',p),
                                        String(v,p),p),Ch(':',p),p),
                                        RegexpID(".",p),p),Ch(')',p),p),p),p))
              ts,
           gapRules ps "" @
             List.map (fun(n,p,es,pos)->(n,p,List.map modify_entity es,
                                         pos)) ps) in
        

        (* Write file to temporary location *)

        let f = open_out !xsgFilename in
        writeXSG f lcon' "" ;
        close_out f
      end ;
    
    (* Preprocess string and save to disc *)

    let position_map = 
    begin
      let f = open_out_bin !fragmentFilename in
      let mode = ref 0 in
      let num = ref "" in
      let typ = ref "" in
      let pos = ref 0 in
      let position_map = ref [] in
      let preprocessor c =
	position_map := !pos :: !position_map ;
	match !mode,c with
          0, '$' -> mode := 1
        | 0, _   -> output_char f 'c' ; output_char f c ; pos := (!pos + 2)
        | 1, '$' -> mode := 0 ; output_string f "c$" ; pos := (!pos + 2)
        | 1, '(' -> mode := 2
        | 1, _   -> mode := 0 ; let s = ("(" ^
                                  (check_typ (make 1 c) "" tenv) ^ ":" ^
                                  (string_of_int (check_num (make 1 c))) ^ ")")
	                        in output_string f s ;
	                           pos := (!pos + (String.length s))
	| 2, _   -> mode := 3 ; num := (make 1 c) ; typ := ""
        | 3, ':' -> mode := 4
        | 3, ')' 
        | 4, ')' -> mode := 0 ; let s = ("(" ^
                                     (check_typ !num !typ tenv) ^ ":" ^
                                     (string_of_int (check_num !num)) ^ ")")
	                        in output_string f s ;
	                           pos := (!pos + (String.length s))
        | 3, _   -> num := (!num ^ (make 1 c))
        | _, _   -> typ := (!typ ^ (make 1 c))
      in
      output_string f (n ^ ":") ;
      pos := (String.length n) + 1 ;
      String.iter preprocessor s ; 
      position_map := !pos :: !position_map ;
      close_out f ;
      position_map
    end in

    (* Call XSugar *)

    let xmlFilename = Filename.temp_file "shape" ".xml" in
    if (Sys.command ("java -jar xsugar-all.jar -q -mo " ^ !xsgFilename ^ " " ^
                     !fragmentFilename ^ " > " ^ xmlFilename)) <> 0 then
      begin
	let f = 
	  try
            open_in xmlFilename 
	  with
            Sys_error _ -> raise (Failure "XSugar failed to create a file")
	in
	if (in_channel_length f) > 0 then
	  let first_line = input_line f in
	  let e =
	  if (String.sub first_line 0 28) = "*** parse error at character" then
	    (* Magic conversion - don't attempt to figure out how this works *)
	    let pos = (List.length (List.filter ((>) (List.find ((>=)
		       (int_of_string (String.sub first_line 29 
		       ((String.index_from first_line 29 ' ') - 29))))
		       !position_map)) !position_map)) + 1 in 
	    "Parse error near position " ^ string_of_int pos ^ " in:\n" ^ s
	  else
	    "XSugar failed with the message:\n" ^ first_line ^
	    if (String.sub first_line ((String.length first_line)-25) 25)
		= "cannot derive any strings" then
	      "\nPerhaps the language of inter-token space ($) in the\n" ^
	      "target language is empty (#)"
	    else ""
	  in
	  begin
	    close_in f ;
	    begin
              try
		Sys.remove xmlFilename
              with Sys_error _ -> ()
	    end ;
	    raise (Invalid_argument e)
	  end
	else
          begin
	    close_in f ;
	    begin
              try
		Sys.remove xmlFilename
              with Sys_error _ -> ()
	    end ;
	    raise (Failure ("XSugar failed - perhaps xsugar-all.jar " ^
			    "was not found in the Java CLASSSPATH ?"));
          end ;
      end ;
    
    (* Read and parse XSugar result into AST *)

    let f = 
      try
        open_in xmlFilename 
      with
        Sys_error _ -> raise (Failure "XSugar failed to create a file")
    in
      let lexbuf = Lexing.from_channel f in
      begin
        let ast = 
          try 
            InlineAuxParser.ast InlineAuxLexer.token lexbuf
          with 
            Parsing.Parse_error ->
            raise (Failure "Inline parser: syntax error")
          | x -> raise x 
        in
          begin
            close_in f ;
            Sys.remove xmlFilename ; 
            ast
          end
      end
  end
