(***********************************************************************)
(*                                                                     *)
(*                           Shape Lexer                               *)
(*                                                                     *)
(*                 Claus Brabrand, Jacob Andersen                      *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)

{
  open Parser     (** The type token is defined in parser.mli **)
  open Lexing     (**           Useful Functions              **) 

  type mode = File | Constantl | Constantx | Regexp | Charclass | Id

  let mode = ref File

  let reset() = mode := File

  let increment lexbuf =
    let 
      pos = lexbuf.lex_curr_p
    in
      lexbuf.lex_curr_p <- 
      {pos with
        pos_lnum = pos.pos_lnum + 1;
        pos_bol = pos.pos_cnum;
      }

  exception LexError of string * (position * position)

}

let newline    = '\n' | '\r' | "\r\n" 
let space      = [' ' '\t']
let comment    = "//" [^'\n' '\r']*
let identifier = (['A'-'Z' 'a'-'z' '_'] ['A'-'Z' 'a'-'z' '0'-'9' '_']*) as i


(* -- Regular Expressions -------------------------------------------- *)

rule regexp = parse
  |  '.'   { DOT        }
  |  '#'   { HASH       }
  |  "+"   { PLUS       }
  |  "*"   { STAR       }
  |  "?"   { QUESTION   }
  |  "|"   { BAR        }
  |  '('   { LPAR       }
  |  ')'   { RPAR       }
  |  "[^"
       {
         mode := Charclass ;
         LBRACKHAT
       }
  |  '['
       {
         mode := Charclass ;
         LBRACK
       }
  |  ';'
       {
         mode := Constantl ;
         SEMIC
       }
  |  '<'
       {
         mode := Id ;
         LT
       }
  |  space     
       {  
         regexp lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         regexp lexbuf
       }    
  |  comment
       {
         regexp lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ; 
         regexp lexbuf
       }
  |  ('"' ([^ '"' '\n' '\r' '\\'] | ('\\'['n''r''t''\\''\'''"']))*  '"') as s
       {
           STR (Scanf.sscanf s "%S" (fun s->s))
       }
  |  '\\' ([^ '\n' '\r' '\t'] as c)
       {
         if c='n' || c='r' || c='t' then
           CHAR (Scanf.sscanf ("'\\" ^ (String.make 1 c) ^ "'")
                              "%C" (fun c->c))
         else
           CHAR c
       }
  |  _ as c
       { 
         CHAR c
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }


(* -- Charclass (regular expressions) -------------------------------- *)

and charclass = parse
  |  "-"   { MINUS      }
  |  ']'
       {
         mode := Regexp ;
         RBRACK
       }
  |  space     
       {  
         charclass lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         charclass lexbuf
       }    
  |  comment
       {
         charclass lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ;
         charclass lexbuf
       }
  |  '\\' ([^ '\n' '\r' '\t'] as c)
       {
         if c='n' || c='r' || c='t' then
           CHAR (Scanf.sscanf ("'\\" ^ (String.make 1 c) ^ "'")
                              "%C" (fun c->c))
         else
           CHAR c
       }
  |  _ as c
       { 
         CHAR c
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }


(* -- Identifier (regular expressions) ------------------------------- *)

and id = parse
  |  '>'
       {
         mode := Regexp ;
         GT
       }
  |  identifier
       {
         ID i
       }
  |  space     
       {  
         id lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         id lexbuf
       }    
  |  comment
       {
         id lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ;
         id lexbuf
       }
  |  _
       { 
	 raise (LexError(Lexing.lexeme lexbuf,
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf )))
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }

(* -- Constant Language ---------------------------------------------- *)

and constantl = parse
  |  "$"               { DOLLAR     }
  |  '.'               { DOT        }
  |  ';'               { SEMIC      }
  |  ':'               { COLON      }
  |  '='
       {
         mode := Regexp ;
         EQUAL
       }
  |  '}'
       {
         mode := File ;
         RBRACE
       }
  |  ('"' ([^ '"' '\n' '\r' '\\'] | ('\\'['n''r''t''\\''\'''"']))*  '"') as s
       {
         STR (Scanf.sscanf s "%S" (fun s->s))
       }
  |  space     
       {  
         constantl lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         constantl lexbuf
       }    
  |  comment
       {
         constantl lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ;
         constantl lexbuf
       }
  |  identifier
       {
         ID i
       }
  |  _
       { 
	 raise (LexError(Lexing.lexeme lexbuf,
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf )))
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }


(* -- Constant Transformation ---------------------------------------- *)

and constantx = parse
  |  '.'               { DOT        }
  |  ','               { COMMA      }
  |  ';'               { SEMIC      }
  |  '='               { EQUAL      }
  |  "->"              { ARROW      }
  |  '('               { LPAR       }
  |  ')'               { RPAR       }
  |  ']'               { RBRACK     }
  |  "|)"
       {
         mode := File ;
         RBANANA
       }
  |  ('"' ([^ '"' '\n' '\r' '\\'] | ('\\'['n''r''t''\\''\'''"']))*  '"') as s
       {
         STR (Scanf.sscanf s "%S" (fun s->s))
       }
  |  '\''
       {
	 let startPos = lexbuf.lex_start_p in
	 let s = multilinestring "" lexbuf in
	 lexbuf.lex_start_p <- startPos ;
         MLSTR s
       }
  |  space     
       {  
         constantx lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         constantx lexbuf
       }    
  |  comment
       {
         constantx lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ;
         constantx lexbuf
       }
  |  '$' ((['1'-'9']['0'-'9']*) as i)
       {
         ARG (int_of_string i)
       }
  |  identifier
       {
         ID i
       }
  |  _
       { 
	 raise (LexError(Lexing.lexeme lexbuf,
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf )))
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }


(* -- Multi line string ---------------------------------------------- *)

and multilinestring sofar = parse
  | "''"
      {
        multilinestring (sofar ^ "'") lexbuf
      }
  | '\''
      {
        sofar
      }
  | [^ '\'' '\n' '\r']* as s
      {
        multilinestring (sofar ^ s) lexbuf
      }
  |  newline
       {
         increment lexbuf ;
         multilinestring (sofar ^ "\n") lexbuf
       }    
  |  _
       { 
	 raise (LexError(Lexing.lexeme lexbuf,
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf )))
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }



(* -- File ----------------------------------------------------------- *)

and file = parse
  |  "let"             { LET        }
  |  "letx"            { LETX       }
  |  "in"              { IN         }
  |  "src"             { SRC        }
  |  "tgt"             { TGT        }
  |  "idx"             { IDX        }
  |  ','               { COMMA      }
  |  '='               { EQUAL      }
  |  "->"              { ARROW      }
  |  "+"               { PLUS       }
  |  "\\"              { BACKSLASH  }
  |  "|"               { BAR        }
  |  "(|"              { LBANANA    }
  |  '('               { LPAR       }
  |  ')'               { RPAR       }
  |  '['
    {
      mode := Constantx ;
      LBRACK
    }
  |  '{'
    {
      mode := Constantl ;
      LBRACE
    }
  |  eof               { EOF        }
  |  space     
       {  
         file lexbuf
       }    
  |  newline
       {
         increment lexbuf ;
         file lexbuf
       }    
  |  comment
       {
         file lexbuf
       } 
  |  "/*"
       {
         comment 0 lexbuf ;
         file lexbuf
       }
  |  '"' ([^ '"' '\n' '\r']* as s) '"'
       {
         STR s
       }
  |  identifier
       {
         ID i
       }
  |  _
       { 
	 raise (LexError(Lexing.lexeme lexbuf,
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf )))
       }


(* -- Multi-line comments (can be nested) ---------------------------- *)

and comment level = parse
  |  "/*"
       {
         comment (level+1) lexbuf
       }
  |  "*/"
       {
         if level != 0 then
           comment (level-1) lexbuf
       }
  |  newline
       {
         increment lexbuf ;
         comment level lexbuf
       }    
  |  _
       {
         comment level lexbuf
       }
  |  eof
       {
         raise (LexError("End-Of-File",
                         (Lexing.lexeme_start_p lexbuf,
                          Lexing.lexeme_end_p lexbuf)))
       }

{


  let token lexbuf = match !mode with
      File       -> file lexbuf
    | Constantl  -> constantl lexbuf
    | Constantx  -> constantx lexbuf
    | Regexp     -> regexp lexbuf
    | Charclass  -> charclass lexbuf
    | Id         -> id lexbuf

}
