(***********************************************************************)
(*                                                                     *)
(*                       Banana Algebra Parse                          *)
(*                                                                     *)
(*                 Claus Brabrand, Jacob Andersen                      *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)

open Ast

exception ParseError of string * pos
exception FileNotFound of string * pos option

let make_ast parser filename pos =
  let channel = 
    try
      open_in filename 
    with
      Sys_error _ -> raise (FileNotFound (filename,pos))
  in
    let lexbuf = Lexing.from_channel channel
    in
      begin
      let pos = lexbuf.Lexing.lex_curr_p in
      lexbuf.Lexing.lex_curr_p <- {pos with Lexing.pos_fname = filename} ;
      Lexer.reset() ;
      let ast = 
        try 
          parser Lexer.token lexbuf
        with 
          Parsing.Parse_error ->
          raise(ParseError(Lexing.lexeme lexbuf,
                           (Lexing.lexeme_start_p lexbuf,
                            Lexing.lexeme_end_p lexbuf)))
        | x -> raise x 
      in
        begin
          close_in channel ;
          ast
        end
    end

let rec
parse_lexp = function
    Lcon(lcon,pos)          -> Lcon(lcon, pos)
  | Lvar(v,pos)             -> Lvar(v, pos)
  | Llet(v,lexp,lexp',pos)  -> Llet(v,parse_lexp lexp, parse_lexp lexp', pos)
  | Lletx(v,xexp,lexp',pos) -> Lletx(v,parse_xexp xexp, parse_lexp lexp', pos)
  | Lplus(lexp,lexp',pos)   -> Lplus(parse_lexp lexp, parse_lexp lexp', pos)
  | Lminus(lexp,lexp',pos)  -> Lminus(parse_lexp lexp, parse_lexp lexp', pos)
  | Lsrc(xexp,pos)          -> Lsrc(parse_xexp xexp, pos)
  | Ltgt(xexp,pos)          -> Ltgt(parse_xexp xexp, pos)
  | Linclude(s,pos)         -> parse_lexp (make_ast Parser.lexp s (Some pos))

and
parse_xexp = function
    Xcon(xs,t,ls,lt,pos)    -> Xcon(xs,t,parse_lexp ls,parse_lexp lt,pos)
  | Xvar(v,pos)             -> Xvar(v,pos)
  | Xlet(v,lexp,xexp',pos)  -> Xlet(v,parse_lexp lexp, parse_xexp xexp',pos)
  | Xletx(v,xexp,xexp',pos) -> Xletx(v,parse_xexp xexp, parse_xexp xexp',pos)
  | Xplus(xexp,xexp',pos)   -> Xplus(parse_xexp xexp, parse_xexp xexp', pos)
  | Xseq(xexp,xexp',pos)    -> Xseq(parse_xexp xexp, parse_xexp xexp', pos)
  | Xidx(lexp,pos)          -> Xidx(parse_lexp lexp, pos)
  | Xinclude(s,pos)         -> parse_xexp (make_ast Parser.xexp s (Some pos))

and
parse filename = parse_xexp (make_ast Parser.xexp filename None)
