(***********************************************************************)
(*                                                                     *)
(*                       Banana Algebra Reduce                         *)
(*                                                                     *)
(*                 Claus Brabrand, Jacob Andersen                      *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)

open Ast
open Transform
open InlineParser

open List

(* -- Merge & Sort --------------------------------------------------- *)

let rec 
unique cmp xs = match xs with
    [] -> []
  | x::[] -> [x]
  | x::y::xs -> let c = cmp x y
                in if (c = 0) 
                then (unique cmp (x::xs))
                else if (c < 0) 
                     then x::(unique cmp (y::xs))
                     else raise (Failure("unique called on unsorted list"))

let usort cmp xs = unique cmp (sort cmp xs)

let umerge cmp xs ys = unique cmp 
                       (merge cmp (usort cmp xs) (usort cmp ys))

let remove cmp xs ys =
  let rec
    aux xs ys = match xs, ys with
        [], _  -> []
      | xs, [] -> xs
      | x::xs', y::ys' -> if cmp x y < 0 then
                            x :: aux xs' ys
                          else if cmp x y = 0 then
                            aux xs' ys'
                          else
                            aux xs ys'
  in
    aux (usort cmp xs) (usort cmp ys)

let
index_of es e = 
  let rec
    aux es e i = match es with 
          [] -> raise (Failure("index-of: element not in list"))
        | x::es -> if (x = e) 
                   then i
                   else aux es e (i+1)
  in
    aux es e 0

(* -- Environments -------------------------------------------------- *)

let lookup env v = assoc v env

let update env a b = (a,b)::env

(* -- lcon ----------------------------------------------------------- *)

let rec
entity_equal e1 e2 = match (e1,e2) with
    (Es(s1,_1),Es(s2,_2)) -> (s1 = s2)
  | (En(n1,_1),En(n2,_2)) -> (n1 = n2)
  | _ -> false

let rec
cmp_prod (n1,p1,es1,pos1) (n2,p2,es2,pos2) = 
  let 
    cmp = (compare n1 n2)
  in 
    if (cmp = 0) then
      let
        cmp = (compare p1 p2)
      in 
        if (cmp = 0) then
          try
            if (for_all2 entity_equal es1 es2) then
              0
            else
              raise (Invalid_argument "")
          with Invalid_argument _ ->
            raise (Conflict ("Two different productions for '" ^ n1 ^ "." ^
                             p1 ^ "'",pos1,pos2))
        else
          cmp
    else
      cmp

and
reduce_prods ps = usort cmp_prod ps

and
regexp_equal r s = match (r,s) with 
    (Union(r11,r12,_1),Union(r21,r22,_2))       -> ((regexp_equal r11 r21) &&
                                                    (regexp_equal r12 r22))
  | (Concat(r11,r12,_1),Concat(r21,r22,_2))     -> ((regexp_equal r11 r21) &&
                                                    (regexp_equal r12 r22))
  | (Optional(r1,_),Optional(r2,_))             -> (regexp_equal r1 r2)
  | (Star(r1,_1),Star(r2,_2))                   -> (regexp_equal r1 r2)
  | (Plus(r1,_1),Plus(r2,_2))                   -> (regexp_equal r1 r2)
  | (CharClass(cs1,_1),CharClass(cs2,_2))       -> (charclasses_equal cs1 cs2)
  | (NegCharClass(cs1,_1),NegCharClass(cs2,_2)) -> (charclasses_equal cs1 cs2)
  | (Dot(_1),Dot(_2))                           -> true
  | (Empty _1, Empty _2)                        -> true
  | (String(s1,_1),String(s2,_2))               -> (s1 = s2)
  | (RegexpID(v1,_1),RegexpID(v2,_2))           -> (v1 = v2)
  | (Ch(c1,_1),Ch(c2,_2))                       -> (c1 = c2)
  | _                                           -> false

and
charclasses_equal cs1 cs2 = match (cs1,cs2) with
    ([],[])               -> true
  | ((c1::cs1),(c2::cs2)) -> (charclass_equal c1 c2) && 
                             (charclasses_equal cs1 cs2)
  | _                     -> false


and
charclass_equal c1 c2 = match (c1,c2) with
    (CharExp(c1,_1),CharExp(c2,_2))       -> (c1 = c2)
  | (Range(c11,c12,_1),Range(c21,c22,_2)) -> ((c11 = c21) && (c12 = c22))
  | _                                     -> false

and
cmp_term (t1,r1,pos1) (t2,r2,pos2) = 
  let
    cmp = (compare t1 t2)
  in
    if (cmp = 0) then
      if (regexp_equal r1 r2) then
        0
      else
        raise (Conflict ("Two different regular expressions for '" ^ t1 ^
                         "'", pos1, pos2))
    else
      cmp

(* -- topsort terminals -- *)

(* imperative definitions for topsort: *)
and
terms = ref []
and
terms_being_generated = ref []
and
terms_done_generated = ref []

and
find_terminal x (t,r,pos) = (t = x)

and
topsort_regexp (r : regexp) : term list = match r with
    Union(r1,r2,_)  -> (topsort_regexp r1) @ (topsort_regexp r2)
  | Concat(r1,r2,_) -> (topsort_regexp r1) @ (topsort_regexp r2)
  | Optional(r,_)   -> topsort_regexp r
  | Star(r,_)       -> topsort_regexp r
  | Plus(r,_)       -> topsort_regexp r
  | RegexpID(v,_)   -> begin
                         try
                           let t = (find (find_terminal v) !terms)
                           in topsort_term t 
                         with 
                           Not_found -> []
                       end
  | _               -> []

and
topsort_term (t,r,pos) : term list = 
  begin
    if (mem t !terms_done_generated) then
      [] (* t already generated *)
    else
      if (mem t !terms_being_generated) then
        raise (Error ("Cyclic regexp definition '" ^ t ^ "'",pos))
      else
        begin
          terms_being_generated := (t::!terms_being_generated) ;
          let ts' = topsort_regexp r in
          terms_done_generated := (t::!terms_done_generated) ;
          ts' @ [(t,r,pos)]
        end
  end

and
topsort_terms (ts : term list) : term list = match ts with
    []    -> []
  | t::ts -> let t' = (topsort_term t) in
             let ts' = (topsort_terms ts) in
             t' @ ts'

and
topsort (ts : term list) : term list = 
  begin
    terms := ts ;
    terms_being_generated := [] ;
    terms_done_generated := [] ;
    topsort_terms ts
  end

and
reduce_terms ts = topsort (usort cmp_term ts)

and
disjoint_names (ts : term list) (ps : prod list) : unit =
  let cmp = fun (v1,_1) (v2,_2) -> (compare v1 v2) in
  let l1 = usort cmp (map (fun (v,_,p) -> (v,p)) ts) in
  let l2 = usort cmp (map (fun (v, _, _, p) -> (v,p)) ps ) in
  let l = merge cmp l1 l2 in
  let rec report_conflict l = match l with
              [] -> ()
            | _::[] -> ()
            | (v1,p1)::(v2,p2)::l ->
              if (v1 = v2) then 
                raise (Conflict(v1,p1,p2))
              else
                report_conflict ((v2,p2)::l)
  in report_conflict l

and
unify_inter_token_space (w1,pos1) (w2,pos2) = match (w1,w2) with
    (Empty _,_) -> (w2,pos2)
  | (_,Empty _) -> (w1,pos1)
  | _ -> if regexp_equal w1 w2 then
           (w1,pos1)
         else
           (Union(w1,w2,pos1),pos1)

and
lplus (w1,ts1,ps1) (w2,ts2,ps2) =
  let w = unify_inter_token_space w1 w2 in
  let ts = topsort (umerge cmp_term ts1 ts2) in
  let ps = umerge cmp_prod ps1 ps2 in
  begin
    (try
      disjoint_names ts ps ;
    with Conflict (v,p1,p2) ->
      raise (Conflict("'" ^ v ^ "' used as non-terminal identifier in one " ^
             "language and regexp identifier in the other" ,p1,p2))) ;
    (w,ts,ps)
  end

and
lminus (w,ts1,ps1) (_,ts2,ps2) =
  let ts = remove cmp_term ts1 ts2 in
  let ps = remove cmp_prod ps1 ps2 in
  (w,ts,ps)

(* -- xcon ----------------------------------------------------------- *)

and
cmp_tau (ns1,nt1,pos1) (ns2,nt2,pos2) = 
  let
    cmp = (compare ns1 ns2)
  in
    if (cmp = 0) then
      let
        cmp = (compare nt1 nt2) 
      in
        if (cmp = 0) then
          0
        else
          raise (Conflict ("Two different transformation typings for '" ^
                           ns1 ^ "'",pos1,pos2))
    else
      cmp

and
reduce_tau tau = usort cmp_tau tau

and
cmp_xprod (n1,p1,a1,pos1) (n2,p2,a2,pos2) = 
  let rec
    ast_equal a1 a2 = match (a1,a2) with
        (Aast(n1,p1,as1,_1),Aast(n2,p2,as2,_2)) -> ((n1 = n2) && (p1 = p2)) &&
                                                   asts_equal as1 as2
      | (Aarg(i1,_1),Aarg(i2,_2))               -> (i1 = i2)
      | (Aterm(t1,s1,_1),Aterm(t2,s2,_2))       -> ((t1 = t2) && (s1 = s2))
      | (Aliteral(s1,_1),Aliteral(s2,_2))       -> (s1 = s2)
      | _ -> false
  and
    asts_equal as1 as2 = match (as1,as2) with
        ([],[]) -> true
      | (a1::as1,a2::as2) -> ast_equal a1 a2 && asts_equal as1 as2
      | _ -> false
  in let 
    cmp = (compare n1 n2)
  in 
    if (cmp = 0) then
      let
        cmp = (compare p1 p2)
      in
        if (cmp = 0) then
          if (ast_equal a1 a2) then
            0
          else
            raise (Conflict ("Two different X productions for '" ^ n1 ^
                             "." ^ p1 ^ "'", pos1, pos2))
        else
          cmp
    else
      cmp

and
reduce_xprods chi = usort cmp_xprod chi

and
xplus (chi1,tau1,ls1,lt1) (chi2,tau2,ls2,lt2) =
  (umerge cmp_xprod chi1 chi2, umerge cmp_tau tau1 tau2,
   lplus ls1 ls2, lplus lt1 lt2)

and
xseq_leq ((_,ts1,ps1):lcon) (((_,ts2,ps2) as l2):lcon) =
  let rec
    prods ps1 ps2 = match ps1 with
        [] -> ()
      | (n1,p1,es1,pos1)::ps1 ->
          let
            (_,_,es2,pos2) =
              try
                get_language l2 n1 p1
              with
                Invalid_argument s ->
                  raise (Error(s ^ " undefined in subsequent " ^
                                   "transformation", pos1))
          in
            try
              if (for_all2 entity_equal es1 es2) then
                prods ps1 ps2
              else
                raise (Invalid_argument "")
            with Invalid_argument _ ->
              raise (Conflict("'" ^ n1 ^ "." ^ p1 ^ "' defined differently " ^
                              "in subsequent transformation",pos1,pos2))

  in
    prods ps1 ps2


and
xseq_chi chi1 x2 = match chi1 with
    [] -> []
  | (n1,p1,a1,pos1)::chi1 ->
      let
        a = transform x2 a1
      in
        (n1,p1,a,pos1)::xseq_chi chi1 x2

and
xseq_tau tau1 tau2 = match tau1 with
    [] -> []
  | (a,b,pos)::tau1 -> try
                         (a,get_tau tau2 b,pos) :: xseq_tau tau1 tau2
                       with Not_found ->
                         raise (Error("'" ^ b ^ "' untyped in subsequent " ^
                                      "transformation",pos))

and
xseq (chi1,tau1,ls1,lt1) ((chi2,tau2,ls2,lt2) as x2) =
  begin
    xseq_leq lt1 ls2 ; 
    let chi = xseq_chi chi1 x2 in
    let tau = xseq_tau tau1 tau2 in
    (chi, tau, ls1, lt2)
  end

and
names ((_,ts,ps):lcon) : string list =
  usort compare (fold_right 
                  (fun (n,_,a,_) ns -> fold_right 
                                         (function Es(_,_) -> (fun ns->ns) 
                                                 | En(n,_) -> (fun ns->n::ns))
                                         a
                                         (n::ns))
                  ps 
                  (map (fun (v,_,_) -> v) ts))

and
idx lcon pos =
  let rec
    id_entities es i = match es with
      [] -> []
    | Es(s,_)::es -> id_entities es i
    | En(n,_)::es -> Aarg(i,pos) :: (id_entities es (i+1))
  in 
    let rec 
      id_xprod pos (n,p,a,_) = (n,p,Aast(n,p,id_entities a 1,pos),pos)
    in
      (map (id_xprod pos) ((fun (_,_,x)-> x) lcon),
       map (fun n -> (n,n,pos)) (names lcon), lcon, lcon)

(* -- xcon wellformedness -------------------------------------------- *)

and
make_tenv (tau : tau) (es : entity list) (pos : pos) : string list =
  match es with
    []    -> []
  | Es(s,_)::es -> make_tenv tau es pos
  | En(n,_)::es ->
      try
        (get_tau tau n)::make_tenv tau es pos
      with Not_found -> ("*" ^ n ^ "*" ^ (Marshal.to_string pos [])) ::
                        make_tenv tau es pos
  
and
wellformed_chi ((chi,tau,ls,lt):xcon) : xprod list = match chi with
    [] -> []
  | (n,p,a,posx)::chi ->
      begin
        let (_,_,es,posl) =
              try
                get_language ls n p
              with Invalid_argument s ->
                raise (Error("Transformation production " ^ s ^
                             " undefined in source language", posx))
        in
          let tenv = make_tenv tau es posl
          in
            let nl = try
                       get_tau tau n
                     with Not_found ->
                       raise (Error("Source language nonterminal '" ^ n ^
                              "' not found in transformation typing",posx))
            in
              begin
                let rec 
                  parse = function
                      Aast (n,p,a,pos) -> Aast (n,p,map parse a,pos)
                    | Aarg (i,pos)     -> Aarg (i,pos)
                    | Aterm (v,s,pos)  -> Aterm (v,s,pos)
                    | Aliteral (s,pos) -> begin
                                            try
                                              inline_parse tenv lt nl s
                                            with Invalid_argument s ->
                                              raise (Error (s,pos))
                                          end
                in
                  let
                    a' = parse a
                  in
                    check_ast tenv lt nl a' ;
                    (n,p,a',posx)::(wellformed_chi (chi,tau,ls,lt))
              end
      end

and
wellformed_xcon ((chi,tau,ls,lt) : xcon) : xcon =
  ((wellformed_chi (chi,tau,ls,lt)),tau,ls,lt)


(* -- lexp ----------------------------------------------------------- *)

and
reduce_lexp (lenv : (string*lcon) list) (xenv : (string*xcon) list)
            (lexp : lexp) : lcon = match lexp with
    Lcon((w,ts,ps),pos)   -> let ts = reduce_terms ts in
                             let ps = reduce_prods ps in
                             begin
                               (try
                                 disjoint_names ts ps
                               with Conflict (v,p1,p2) ->
                                 raise (Conflict ("'" ^ v ^ "' used as " ^
                                 "both non-terminal and regexp identifier",
                                 p1,p2))) ;
                               (w, ts, ps)
                             end
  | Lvar(v,pos)           -> begin
                               try
                                 lookup lenv v
                               with
                                 Not_found ->
                                   raise (Error("Undefined language: '" ^
                                                v ^ "'", pos))
                             end
  | Llet(v,lexp,lexp',_)  -> let l = reduce_lexp lenv xenv lexp
                             in reduce_lexp (update lenv v l) xenv lexp'
  | Lletx(v,xexp,lexp',_) -> let x = reduce_xexp lenv xenv xexp
                             in reduce_lexp lenv (update xenv v x) lexp'
  | Lplus(lexp,lexp',pos) -> let l = reduce_lexp lenv xenv lexp in
                             let l' = reduce_lexp lenv xenv lexp' in
                             begin
                               try
                                 lplus l l'
                               with
                                 Conflict(s,p1,p2) ->
                                   raise (SourcedConflict ("'+L' conflict: "
                                                           ^ s,pos,p1,p2))
                             end
  | Lminus(lexp,lexp',pos)-> let l = reduce_lexp lenv xenv lexp in
                             let l' = reduce_lexp lenv xenv lexp' in
                             begin
                               try
                                 lminus l l'
                               with
                                 Conflict(s,p1,p2) ->
                                   raise (SourcedConflict ("'\\' conflict: "
                                                           ^ s,pos,p1,p2))
                             end
  | Lsrc(xexp,_)          -> let (chi,tau,ls,lt) = reduce_xexp lenv xenv xexp
                             in ls
  | Ltgt(xexp,_)          -> let (chi,tau,ls,lt) = reduce_xexp lenv xenv xexp
                             in lt
  | Linclude(s,_)         -> raise (Failure("L-include should have been " ^
                                            "inlined during normalization"))

(* -- xexp ----------------------------------------------------------- *)

and
reduce_xexp (lenv : (string*lcon) list) (xenv : (string*xcon) list)
            (xexp : xexp) : xcon = match xexp with
    Xcon(chi,tau,ls,lt,_) -> wellformed_xcon (reduce_xprods chi,
                                              reduce_tau tau,
                                              reduce_lexp lenv xenv ls,
                                              reduce_lexp lenv xenv lt)
  | Xvar(v,pos)           -> begin
                               try
                                 lookup xenv v
                               with
                                 Not_found ->
                                   raise (Error("Undefined transformation: '"
                                                ^ v ^ "'", pos))
                             end
  | Xlet(v,lexp,xexp',_)  -> let l = reduce_lexp lenv xenv lexp
                             in reduce_xexp (update lenv v l) xenv xexp'
  | Xletx(v,xexp,xexp',_) -> let x = reduce_xexp lenv xenv xexp
                             in reduce_xexp lenv (update xenv v x) xexp'
  | Xplus(xexp,xexp',pos) -> let x = reduce_xexp lenv xenv xexp in
                             let x' = reduce_xexp lenv xenv xexp' in
                             begin
                               try
                                 xplus x x'
                               with
                                 Conflict(s,p1,p2) ->
                                   raise (SourcedConflict ("'+X' conflict: "
                                                           ^ s,pos,p1,p2))
                             end
  | Xseq(xexp,xexp',pos)  -> let x = reduce_xexp lenv xenv xexp in
                             let x' = reduce_xexp lenv xenv xexp' in
                             begin
                               try
                                 xseq x x'
                               with
                                 Error(s,p) ->
                                   raise (SourcedError ("Composition error: " ^
                                                        s,pos,p))
                               | Conflict(s,p1,p2) ->
                                   raise (SourcedConflict ("Composition " ^
                                          "conflict: " ^ s,pos,p1,p2))
                             end
  | Xidx(lexp,pos)        -> let l = reduce_lexp lenv xenv lexp in
                             idx l pos
  | Xinclude(s,_)         -> raise (Failure("X-include should have been " ^
                                            "inlined during normalization"))

and
reduce xexp = reduce_xexp [] [] xexp
